/* Bonk - lossy/lossless audio compressor
   Copyright (C) 2001  Paul Francis Harrison

   Modified by Robert Kausch for use with the Bonk plugin for Winamp.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  675 Mass Ave, Cambridge, MA 02139, USA.

  The author may be contacted at:
    pfh@csse.monash.edu.au
  or
    3 Currajong St., Oakleigh East, 3166, Melbourne, Australia

  See also
    http://yoyo.cc.monash.edu.au/~pfh/
*/

#include "utility.h"
#include "codec.h"

BONKdecoder::BONKdecoder()
{
}

BONKdecoder::~BONKdecoder()
{
	end();
}

void BONKdecoder::begin(InStream *_f_in)
{
	f_in		= _f_in;
	info_pos	= 0;
	data_pos	= 0;
	isBonk		= false;
	id3size		= 0;

	// Looking for META tag:

	f_in->Seek(f_in->Size() - 4);

	if (f_in->InputNumber(4) == 1635018093) // 'meta' string
	{
		f_in->Seek(f_in->Size() - 8);

		int tagSize = f_in->InputNumber(4);

		f_in->Seek(f_in->Size() - tagSize + 6);

		for (int i = 0; i < (tagSize - 14) / 8; i++)
		{
			switch (f_in->InputNumber(4))
			{
				case 1802399586: // 'bonk' string
					data_pos = f_in->InputNumber(4) + 22;
					break;
				case 1868983913: // 'info' string
					info_pos = f_in->InputNumber(4) + 6;
					break;
			}
		}

		f_in->Seek(data_pos - 22);

		if (f_in->InputNumber(1) == 0 &&
		    f_in->InputNumber(4) == 1263423298 &&
		    f_in->InputNumber(1) == 0) isBonk = true;
	}
	else
	{
		vector<char>	buffer;

		for(int i = 0; i < 6; i++) buffer.push_back(f_in->InputNumber(1));

		for(int i = 0; i < f_in->Size() - 6; i++)
		{
			if (buffer[buffer.size()-6] == 0   &&
			    buffer[buffer.size()-5] == 'I' &&
			    buffer[buffer.size()-4] == 'N' &&
			    buffer[buffer.size()-3] == 'F' &&
			    buffer[buffer.size()-2] == 'O' &&
			    buffer[buffer.size()-1] == 0)
			{
				info_pos = f_in->GetPos();
				break;
			}

			buffer.push_back(f_in->InputNumber(1));
		}

		f_in->Seek(0);

		for(int i = 0; i < 6; i++) buffer.push_back(f_in->InputNumber(1));

		for(int i = 0; i < f_in->Size() - 6; i++)
		{
			if (buffer[buffer.size()-6] == 0   &&
			    buffer[buffer.size()-5] == 0   &&
			    buffer[buffer.size()-4] == 'I' &&
			    buffer[buffer.size()-3] == 'D' &&
			    buffer[buffer.size()-2] == '3' &&
			    buffer[buffer.size()-1] == 3)
			{
				id3size = -1;
				break;
			}

			buffer.push_back(f_in->InputNumber(1));
		}

		if (id3size != 0)
		{
			f_in->InputNumber(1);	// minor version byte
			f_in->InputNumber(1);	// flags byte

			id3size = 0;

			for (int i = 0; i < 4; i++) id3size = (id3size << 7) | f_in->InputNumber(1);

			id3tag = new unsigned char [id3size + 10];

			f_in->RelSeek(-10);

			for (int i = 0; i < id3size; i++) id3tag[i] = f_in->InputNumber(1);
		}

		f_in->Seek(0);

		for(int i = 0; i < 6; i++) buffer.push_back(f_in->InputNumber(1));

		for(int i = 0; i < f_in->Size() - 6; i++)
		{
			if (buffer[buffer.size()-6] == 0   &&
			    buffer[buffer.size()-5] == 'B' &&
			    buffer[buffer.size()-4] == 'O' &&
			    buffer[buffer.size()-3] == 'N' &&
			    buffer[buffer.size()-2] == 'K' &&
			    buffer[buffer.size()-1] == 0)
			{
				isBonk = true;
				break;
			}

			buffer.push_back(f_in->InputNumber(1));
		}
	}

	if (!isBonk) return;

	length			= uint32(f_in->InputNumber(4));
	rate			= uint32(f_in->InputNumber(4)); 
	channels		= uint8(f_in->InputNumber(1));
	lossless		= uint8(f_in->InputNumber(1));
	mid_side		= uint8(f_in->InputNumber(1));
	n_taps			= uint16(f_in->InputNumber(2));
	down_sampling		= uint8(f_in->InputNumber(1));
	samples_per_packet	= uint16(f_in->InputNumber(2));

	data_pos = f_in->GetPos();

	predictor.init(n_taps);
	predictor_initer.resize(channels);

	for(int i = 0; i < channels; i++)
	{
		predictor_initer[i].resize(n_taps);

		for(int j = 0; j < n_taps; j++)
		{
			predictor_initer[i][j] = 0;
		}
	}

	bit_in.setup(f_in);

	length_remaining	= length;
	currentPacket		= 0;

	currentBitrate		= int(double(f_in->Size() - data_pos) * 8 / length * rate * channels / 1000);
}

void BONKdecoder::end()
{
	if (id3size > 0)
	{
		id3size = 0;

		delete [] id3tag;
	}
}

void BONKdecoder::skip_packets(int packets)
{
	if (!isBonk) return;
	if (packets == 0) return;

	if (info_pos == 0)
	{
		vector<int>	 samples;

		if (packets < 0)
		{
			f_in->Seek(data_pos);

			bit_in.setup(f_in);

			length_remaining	= length;
			packets			= currentPacket + packets;
			currentPacket		= 0;
		}

		for (int i = 0; i < packets; i++) read_packet(samples);

		return;
	}

	currentPacket += packets;

	f_in->Seek(info_pos + currentPacket * 5);

	int newpos		= f_in->InputNumber(4) + data_pos;
	bit_in.bit_no		= f_in->InputNumber(1);
	length_remaining	= length_remaining - packets * samples_per_packet * down_sampling * channels;

	if (currentPacket != 0)
	{
		f_in->Seek(info_pos + (currentPacket - 1) * 5);

		currentBitrate	= int(double(newpos - f_in->InputNumber(4)) * 8 / (samples_per_packet * down_sampling) * rate / 1000);
	}

	f_in->Seek(newpos - 1);
 
	bit_in.byte		= f_in->InputNumber(1);

	if (currentPacket == 0) bit_in.byte = 0;
}

void BONKdecoder::read_packet(vector<int> &samples)
{
	if (!isBonk) return;

	currentPacket++;

	int cpos = f_in->GetPos();

	samples.resize(samples_per_packet * down_sampling * channels);

	vector<int> input_samples(samples_per_packet);

	read_list(predictor.k, false, bit_in);

	predictor.dequantize();

	int quant = (lossless ? 1 : bit_in.read_uint(16) * sample_factor);
 
	for(int channel = 0; channel < channels; channel++)
	{
		int *sample = &(samples[channel]);

		predictor.state = predictor_initer[channel];
		predictor.init_state();

		read_list(input_samples, true, bit_in);
      
		for(int i = 0; i < samples_per_packet; i++)
		{
			for(int j = 0; j < down_sampling - 1; j++)
			{
				*sample = predictor.advance_by_error(0);
				sample += channels;
			}

			*sample = predictor.advance_by_error(input_samples[i] * quant);
			sample += channels;
		}

		for(int i = 0; i < n_taps; i++)
		{
			predictor_initer[channel][i] = samples[samples.size() - channels + channel - i * channels];
		}
	}

	if (mid_side)
	{
		for(int i = 0; i < (int) samples.size(); i += channels)
		{
			samples[i + 1]	+= shift(samples[i], 1);
			samples[i]	-= samples[i + 1];
		}  
	}

	if (!lossless)
	{
		for(int i = 0; i < (int) samples.size(); i++)
		{
			samples[i] = shift(samples[i], sample_shift);
		}
	}

	if (length_remaining < (int) samples.size())
	{
		samples.resize(length_remaining);
		length_remaining = 0;
	}
	else
	{
		length_remaining -= samples.size();
	}

	currentBitrate = int(double(f_in->GetPos() - cpos) * 8 / (samples_per_packet * down_sampling) * rate / 1000);
}

int BONKdecoder::get_id3tag_size()
{
	return id3size;
}

void BONKdecoder::get_id3tag(unsigned char *buffer)
{
	if (id3size == 0) return;

	for (int i = 0; i < id3size; i++) buffer[i] = id3tag[i];
}

void BONKdecoder::set_id3tag(unsigned char *buffer, int size)
{
	if (size == 0) return;

	id3tag = new unsigned char [size];

	for (int i = 0; i < size; i++) id3tag[i] = buffer[i];

	id3size = size;
}
