/* Bonk plugin for Winamp - by Robert Kausch 2001-2002
 *
 * TODO: implement info dialog
 *	 interpret artist/title tags
 *	 use SMOOTHThreads and SMOOTHStrings
 *
 *	 Paul: fix bugs in the bonk encoder/decoder
 */

#define __THROW_BAD_ALLOC exit(1)

#include <windows.h>
#include <math.h>
#include <smoothdll.h>

#include "fileinfo.h"
#include "codec.h"
#include "in.h"

void SMOOTH::AttachDLL()
{
}

void SMOOTH::DetachDLL()
{
}

extern "C" {

#define WM_WA_MPEG_EOF WM_USER+2

#define BPS 16

extern In_Module	 mod;			// the output module (declared near the bottom of this file)
char			 lastfn[MAX_PATH];	// currently playing file (used for getting info on the current file)
int			 decode_pos_ms;		// current decoding position, in milliseconds
int			 paused;		// are we paused?
int			 seek_needed;		// if != -1, it is the point that the decode thread should seek to, in ms.
int16			*sample_buffer = NULL;	// sample buffer

BONKdecoder		 deco;			// decoder
InStream		*f_in = NULL;		// input file handle

int			 killDecodeThread = 0;	// the kill switch for the decode thread
HANDLE			 hthread = (HANDLE) -1;	// the handle to the decode thread

int			 get_samples(int16 *);
DWORD WINAPI __stdcall	 DecodeThread(void *);	// the decode thread procedure

void config(HWND hwndParent)
{
	SMOOTH::MessageBox("There's no configuration for this plugin.", "Bonk plugin for Winamp", MB_OK, IDI_HAND);
}

void about(HWND hwndParent)
{
	SMOOTH::MessageBox("Bonk plugin version 0.6\n\nCopyright (C) Robert Kausch 2001-2002\nCopyright (C) for the Bonk algorithms Paul Harrison 2001", "About Bonk plugin", MB_OK, IDI_INFORMATION);
}

void init()
{
	/* any one-time initialization goes here (configuration reading, etc) */
}

void quit()
{
	/* one-time deinit, such as memory freeing */
}

int isourfile(char *filename)
{
	return 0;
} 

int play(char *filename) 
{
	int	 maxlatency;
	int	 thread_id;

	f_in = new InStream(STREAM_FILE, filename);

	deco.begin(f_in);

	if (!deco.isBonk)
	{
		delete f_in;

		SMOOTH::MessageBox("This is not a Bonk file!", "Bonk error", 0, IDI_HAND);

		return 1;
	}

	sample_buffer	= new int16 [deco.samples_per_packet * deco.down_sampling * deco.channels];

	strcpy(lastfn, filename);

	paused		= 0;
	decode_pos_ms	= 0;
	seek_needed	= -1;

	maxlatency = mod.outMod->Open(deco.rate, deco.channels, BPS, -1, -1);

	mod.SetInfo(deco.currentBitrate, deco.rate / 1000, deco.channels, 1);

	// initialize vis stuff
	mod.SAVSAInit(maxlatency, deco.rate);
	mod.VSASetInfo(deco.rate, deco.channels);

	mod.outMod->SetVolume(-666); // set the output plug-ins default volume

	killDecodeThread = 0;
	hthread = (HANDLE) CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) DecodeThread, (void *) &killDecodeThread, 0, (DWORD *) &thread_id);
	
	return 0; 
}

void pause()
{
	paused = 1;

	mod.outMod->Pause(1);
}

void unpause()
{
	paused = 0;

	mod.outMod->Pause(0);
}

int ispaused()
{
	return paused;
}

void stop()
{ 
	if (hthread != INVALID_HANDLE_VALUE)
	{
		killDecodeThread = 1;

		if (WaitForSingleObject(hthread, INFINITE) == WAIT_TIMEOUT)
		{
			SMOOTH::MessageBox("Error killing decode thread!", "Bonk error", 0, IDI_HAND);
			TerminateThread(hthread, 0);
		}

		CloseHandle(hthread);

		hthread = INVALID_HANDLE_VALUE;
	}

	mod.outMod->Close();

	mod.SAVSADeInit();

	deco.end();

	delete f_in;

	f_in = NULL;

	delete [] sample_buffer;

	lastfn[0] = 0;
}

int getlength()
{ 
	return int(double(deco.length) / (deco.rate * deco.channels) * 1000);
}

int getoutputtime()
{ 
	return decode_pos_ms + (mod.outMod->GetOutputTime() - mod.outMod->GetWrittenTime());
}

void setoutputtime(int time_in_ms)
{ 
	seek_needed = time_in_ms;
}

void setvolume(int volume)
{
	mod.outMod->SetVolume(volume);
}

void setpan(int pan)
{
	mod.outMod->SetPan(pan);
}

int infoDlg(char *filename, HWND hwnd)
{
	fileInfo	*infoDlg = new fileInfo(lastfn);

	infoDlg->ShowDialog();

	delete infoDlg;

	return 0;
}

void getfileinfo(char *filename, char *title, int *length_in_ms)
{
	if (filename == NULL || *filename == 0)  // currently playing file
	{
		if (length_in_ms) *length_in_ms = getlength();

		if (title) 
		{
			char *p = lastfn + strlen(lastfn);

			while (*p != '\\' && p >= lastfn) p--;

			strcpy(title, ++p);

			title[strlen(title) - 5] = 0;
		}
	}
	else // some other file
	{
		if (length_in_ms)
		{
			InStream	*file;
			BONKdecoder	*fdeco;

			*length_in_ms = -1000;

			fdeco = new BONKdecoder();
			file = new InStream(STREAM_FILE, filename);

			fdeco->begin(file);

			*length_in_ms = int(double(fdeco->length) / (fdeco->rate * fdeco->channels) * 1000);

			fdeco->end();

			delete fdeco;
			delete file;
		}

		if (title) 
		{
			char *p = filename + strlen(filename);

			while (*p != '\\' && p >= filename) p--;

			strcpy(title, ++p);

			title[strlen(title) - 5] = 0;
		}
	}
}

void eq_set(int on, char data[10], int preamp) 
{ 
	// most plug-ins can't even do an EQ anyhow. I'm working on writing
	// a generic PCM EQ, but it looks like it'll be a little too CPU 
	// consuming to be useful :)
}

int get_samples(int16 *buf)
{
	if (!deco.length_remaining) return 0;

	vector<int>	 samples;

	deco.read_packet(samples);

	for(unsigned int i = 0; i < samples.size(); i++)
	{
		if (samples[i] > 32767)		buf[i] =  32767;
		else if (samples[i] < -32768)	buf[i] = -32768;
		else				buf[i] =  samples[i];
	}

	return samples.size();
}

DWORD WINAPI __stdcall DecodeThread(void *endDecodeThread)
{
	int done = 0;

	while (!*((int *) endDecodeThread))
	{
		if (seek_needed != -1)
		{
			int	 n_packets = int(double(seek_needed - decode_pos_ms) / (double(deco.samples_per_packet * deco.down_sampling) / deco.rate * 1000));

			decode_pos_ms = int(double(deco.samples_per_packet * deco.down_sampling) / deco.rate * 1000 * (deco.currentPacket + n_packets));
			seek_needed = -1;
			done = 0;

			mod.outMod->Flush(decode_pos_ms);

			deco.skip_packets(n_packets);

			mod.SetInfo(deco.currentBitrate, deco.rate / 1000, deco.channels, 1);
		}

		if (done)
		{
			mod.outMod->CanWrite();

			if (!mod.outMod->IsPlaying())
			{
				PostMessage(mod.hMainWindow, WM_WA_MPEG_EOF, 0, 0);
				return 0;
			}

			Sleep(10);
		}
		else if (mod.outMod->CanWrite() >= ((deco.samples_per_packet * deco.down_sampling * deco.channels * (BPS / 8)) << (mod.dsp_isactive() ? 1 : 0)))
		{
			int	 samples_size = get_samples(sample_buffer);

			if (!samples_size)
			{
				done = 1;
			}
			else
			{
				mod.SetInfo(deco.currentBitrate, deco.rate / 1000, deco.channels, 1);
				mod.SAAddPCMData((char *) sample_buffer, deco.channels, BPS, decode_pos_ms);
				mod.VSAAddPCMData((char *) sample_buffer, deco.channels, BPS, decode_pos_ms);
				decode_pos_ms = int(double(deco.samples_per_packet * deco.down_sampling) / deco.rate * 1000 * deco.currentPacket);

				if (mod.dsp_isactive()) samples_size = mod.dsp_dosamples(sample_buffer, samples_size * 2 / deco.channels / (BPS / 8), BPS, deco.channels, deco.rate) * (deco.channels * (BPS / 8)) / 2;

				mod.outMod->Write((char *) sample_buffer, samples_size * 2);
			}
		}
		else
		{
			Sleep(20);
		}
	}

	return 0;
}

In_Module mod = 
{
	IN_VER,
	"Bonk plugin v0.6 "

#ifdef __alpha
	"(AXP)",
#else
	"(x86)",
#endif

	0,					// hMainWindow
	0,					// hDllInstance
	"Bonk\0Bonk Audio File (*.bonk)\0"
	,
	1,					// is_seekable
	1,					// uses output
	config,
	about,
	init,
	quit,
	getfileinfo,
	infoDlg,
	isourfile,
	play,
	pause,
	unpause,
	ispaused,
	stop,
	
	getlength,
	getoutputtime,
	setoutputtime,

	setvolume,
	setpan,

	0, 0, 0, 0, 0, 0, 0, 0, 0,		// vis stuff

	0, 0,					// dsp

	eq_set,

	NULL,					// setinfo

	0					// out_mod
};

__declspec( dllexport ) In_Module * winampGetInModule2()
{
	return &mod;
}

} /* extern "C" */
