#include "utility.h"
#include <vector>
#include <iolib-cxx.h>

// Number of bits required to store a value:

int bits_to_store(uint32 i) {
  int result = 0;
  while(i) {
    result++;
    i >>= 1;
  }
  return result;
}

// Coder/decoder for lists of small signed ints:

void read_list(vector<int> &list,bool base_2_part, bitstream_in &in) {
  //Read a list of ints as encoded by "write_list"

  //Read base 2 part:

  int low_bits = (base_2_part ? in.read_uint(4) : 0);

  for(int i=0;i<(int)list.size();i++) 
    list[i] = in.read_uint(low_bits);
    
  // Decode bitstream:

  int n_zeros = 0;
  int step = 256;
  bool dominant = false;
  vector<uint8> bits;
  while(n_zeros < (int)list.size()) {
    int steplet = step>>8;
  
    if (!in.read()) {
      for(int i=0;i<steplet;i++)
        bits.push_back(dominant);
      
      if (!dominant)
        n_zeros += steplet;
	
      step += step / adapt_level;
    } else {
      int actual_run = in.read_uint_max(steplet-1);
      
      for(int i=0;i<actual_run;i++)
        bits.push_back(dominant);
      bits.push_back(!dominant);

      if (!dominant)
        n_zeros += actual_run;
      if (dominant)
        n_zeros++;
    
      step -= step / adapt_level;
    }

    if (step < 256) {
      step = 65536 / step;
      dominant = !dominant;
    }
  }

  // Reconstruct (unsigned) values:

  n_zeros = 0;
  int pos   = 0;
  int level = 0;
  for(int i=0;n_zeros<(int)list.size();i++) {
    for(;;) {
      if (pos >= (int)list.size()) {
        pos = 0;
	level += 1<<low_bits;
     }

      if (list[pos] >= level)
        break;

      pos++;
    }
      
    if (bits[i])
      list[pos] += 1<<low_bits;
    else
      n_zeros++;

    pos++;
  }

  // Read signs:

  for(int i=0;i<(int)list.size();i++)
    if (list[i] && in.read())
      list[i] = -list[i];
}
